#define _GNU_SOURCE
#define _FILE_OFFSET_BITS 64
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <utime.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>
#include <libintl.h>
#include <errno.h>

#define MYNAME "mailcount"
#define VERSION "0.07"

const char *usage =
	"\n"
	MYNAME ": prints how many messages you have\n"
	"\n"
	"Usage: " MYNAME " [-d] mailbox [mailbox mailbox ...]\n"
	"       " MYNAME " --help\n"
	"       " MYNAME " --version\n"
	"\n"
	"For example if you have $MAIL set, try ``" MYNAME " $MAIL'',\n"
	"otherwise try ``" MYNAME " /var/mail/your_login_name''.\n"
	"\n"
	"Written by Egmont Koblinger <egmont@fazekas.hu>\n"
	"\n";

const char *version =
	MYNAME " version " VERSION "\n";

enum { notspecial, empty, from_space, from_colon, status };


int linetype (const char *line)
{
	if (line[0] == '\n') return empty;
	if (!strncmp(line, "From ", 5)) return from_space;
	if (!strncmp(line, "From: ", 6)) return from_colon;
	if (!strncmp(line, "Status: ", 8)) return status;
	return notspecial;
}

int mailcount (const char *file, int *All, int *New, int *Unread)
{
	struct stat mailboxstat;
	struct timeval tv[2];
	int fd;
	char *p, *end;
	size_t l;
	int internal, s, r, head, t;

	setlocale(LC_ALL, "");
	bindtextdomain("mailcount", "/usr/share/locale");
	textdomain("mailcount");

	fd = open(file, O_RDONLY);
	if (fd == -1) {
		fprintf(stderr, MYNAME ": unable to open %s: %s\n", file, strerror(errno));
		return 1;
	}
	if (fstat(fd, &mailboxstat)) {
		fprintf(stderr, MYNAME ": unable to fstat %s: %s\n", file, strerror(errno));
		close(fd);
		return 1;
	}
	if (!S_ISREG(mailboxstat.st_mode)) {
		fprintf(stderr, MYNAME ": %s is not a regular file\n", file);
		close(fd);
		return 1;
	}
	l = mailboxstat.st_size;
	if (l == 0) {
		close(fd);
		return 0;
	}
	p = mmap(NULL, l, PROT_READ, MAP_SHARED, fd, 0);
	if (p == MAP_FAILED) {
		fprintf(stderr, MYNAME ": unable to mmap %s: %s\n", file, strerror(errno));
		close(fd);
		return 1;
	}
	end = p+l;

	if (p[l-1] != '\n') {
		fprintf(stderr, MYNAME ": invalid mailbox %s: doesn't end with newline\n", file);
		munmap(p, l);
		close(fd);
		return 1;
	}
	if (linetype(p) != from_space) {
		fprintf(stderr, MYNAME ": invalid mailbox %s: doesn't start with From\n", file);
		munmap(p, l);
		close(fd);
		return 1;
	}

	head = 0;
	internal = s = r = 0;

	while (p != end) {
		t = linetype(p);
		if (head) {
			if (t == from_colon) {
				if (!strncmp(p+6, "Mail System Internal Data", 25))
					internal = 1;
				}
			if (t == status) {
				s = 1;
				p += 8;
				while (*p != '\n') if (*p++ == 'R') r = 1;
				}
			if (t == empty) {
				head = 0;
				if (!internal) {
					(*All)++;
					if (!s) (*New)++;
					if (s && !r) (*Unread)++;
				}
				internal = s = r = 0;
			}
		}
		else {
			if (t == from_space) {
				head = 1;
			}
		}
		while (*p++ != '\n');
	}

	if (head) {
		fprintf(stderr, MYNAME ": invalid mailbox %s: ends in message header\n", file);
		munmap(p, l);
		close(fd);
		return 1;
	}
	munmap(p, l);
	tv[0].tv_sec = mailboxstat.st_atim.tv_sec;
	tv[0].tv_usec = mailboxstat.st_atim.tv_nsec / 1000;
	if (fstat(fd, &mailboxstat)) {
		fprintf(stderr, MYNAME ": unable to fstat %s: %s\n", file, strerror(errno));
		close(fd);
		return 1;
	}
	tv[1].tv_sec = mailboxstat.st_mtim.tv_sec;
	tv[1].tv_usec = mailboxstat.st_mtim.tv_nsec / 1000;
	close(fd);
	utimes(file, tv);
	return 0;
}

int main (int argc, char *argv[])
{
	int args;
	int All, New, Unread, Read;
	int digits_only = 0;

	if (argc < 2) {
		fprintf(stderr, "%s", usage);
		return 1;
	}
	if (!strcmp(argv[1], "--help")) {
		printf("%s", usage);
		return 0;
	}
	if (!strcmp(argv[1], "--version")) {
		printf("%s", version);
		return 0;
	}
	while (argv[1] != NULL && !strcmp(argv[1], "-d")) {
		digits_only = 1;
		argv++;
		argc--;
	}
	if (argc < 2) {
		fprintf(stderr, "%s", usage);
		return 1;
	}

	All = New = Unread = 0;
	args = argc-1;
	while (args) {
		if (mailcount(argv[args], &All, &New, &Unread)) return 1;
		args--;
	}
	Read = All-New-Unread;

	if (digits_only) {
		printf(gettext("%d %d %d\n"), New, Unread, Read);
	} else {
		if (!New) {
			if (!Unread) {
				if (!Read) {
					printf(
					  gettext(
					    "You have no messages.\n"
					    )
					  );
				} else {
					printf(
					  ngettext(
					    "You have %d read message.\n",
					    "You have %d read messages.\n",
					    Read),
					  Read);
				}
			} else {
				if (!Read) {
					printf(
					  ngettext(
					    "You have %d unread message.\n",
					    "You have %d unread messages.\n",
					     Unread),
					   Unread);
				} else {
					printf(
					  ngettext(
					    "You have %d unread and %d read message.\n",
					    "You have %d unread and %d read messages.\n",
					    Read),
					  Unread, Read);
				}
			}
		} else {
			if (!Unread) {
				if (!Read) {
					printf(
					  ngettext(
					    "You have %d new message.\n",
					    "You have %d new messages.\n",
					    New),
					  New);
				} else {
					printf(
					  ngettext(
					    "You have %d new and %d read message.\n",
					    "You have %d new and %d read messages.\n",
					    Read),
					  New, Read);
				}
			} else {
				if (!Read) {
					printf(
					  ngettext(
					    "You have %d new and %d unread message.\n",
					    "You have %d new and %d unread messages.\n",
					    Unread),
					  New, Unread);
				} else {
					printf(
					  ngettext(
					    "You have %d new, %d read and %d unread message.\n",
					    "You have %d new, %d read and %d unread messages.\n",
					    Unread),
					  New, Read, Unread);
				}
			}
		}
	}
	return 0;
}
